;;;; web-unity.scm
;
; Copyright (c) 2007, Peter Bex
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.

;; Ensure we load all extensions needed here, even if only
;; a few subsystems need it so there are no unexpected
;; environment differences between unity subsystems.
;; Look into if/how this can be done using EXPORT etc.
(use regex-case http-utils posix)

(define wu:params (make-parameter #f))
(define wu:cookies (make-parameter #f))
(define wu:query-params (make-parameter #f))
(define wu:path-info (make-parameter #f))
(define wu:script-name (make-parameter #f))
(define wu:script-filename (make-parameter #f))
(define wu:request-method (make-parameter #f))
(define wu:response-headers (make-parameter #f))
(define wu:response-code (make-parameter '(200 . "OK")))
(define wu:headers-sent (make-parameter #f))

;; Basically copied from Spiffy
(define (htmlize str)
  (string-translate* str
		     '(("<" . "&lt;")
		       (">" . "&gt;")
		       ("\"" . "&quot;")
		       ("&" . "&amp;"))))

(define generate-error-message
  (let ()
    (define exn-message (condition-property-accessor 'exn 'message "(no message)"))
    (define exn-location (condition-property-accessor 'exn 'location "(unknown location)"))
    (define exn-arguments (condition-property-accessor 'exn 'arguments '()))
    (define exn? (condition-predicate 'exn))
    (lambda (exn)
      (let ((chain (with-output-to-string print-call-chain)))
	(with-output-to-string
	  (lambda ()
	    (if (exn? exn)
		(begin
		  (display "<h2>Error:")
		  (and-let* ([loc (exn-location exn)])
			    (printf " (<em>~A</em>)" (htmlize (->string loc))))
		  (printf "</h2><h3>~A</h3>" (htmlize (exn-message exn)))
		  (unless (null? (exn-arguments exn))
			  (printf "<ul>")
			  (for-each
			   (lambda (a)
			     (##sys#with-print-length-limit 120 (lambda () (printf "<li>~S</li>" (htmlize (->string a))))))
			   (exn-arguments exn))
			  (printf "</ul>"))
		  (printf "<pre>~a</pre>" (htmlize chain)))
		(begin
		  (##sys#with-print-length-limit
		   120
		   (lambda ()
		     (printf "<hr /><strong>Uncaught exception:</strong> ~S" exn)))))))))))

(define wu:exception-handler
  (make-parameter (lambda (exn)
		    (printf
		     "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\"><html lang=\"en\"><head><title>Error!</title></head><body><h3>Sorry, your request could not be handled.</h3>~A</body></html>"
		     (generate-error-message exn)))))

(define (wu:invoke-dispatcher dispatcher)
  (parameterize ((wu:headers-sent #f)
		 (wu:cookies (get-request-cookies)))
   (handle-exceptions exn (begin (unless (wu:headers-sent)
					 (parameterize ((wu:response-code '(500 . "Internal server error")))
					   (wu:write-headers)))
				 ((wu:exception-handler) exn))
		      (dispatcher))))

(load-verbose #f)  ;; No "loading xyz.scm..." in output

; Add http/1.1 chunking to an existing output port
(define (wu:chunked-output-port p)
  (make-output-port
   (lambda (data)
     (fprintf p "~X\r\n~A\r\n" (string-length data) data))
   void))

;; From Spiffy
(define (wu:single-header? name)
  (not (member name '("Set-Cookie") string-ci=?)))

(define (wu:set-header! name value)
  (if (wu:single-header? name)
      (let* ([rh (wu:response-headers)]
	     [a (assoc name rh string-ci=?)] )
	(if a
	    (set-cdr! a value)
	    (wu:response-headers (alist-cons name value rh)) ) )
      (wu:response-headers (alist-cons name value (wu:response-headers)))))

;;; Cookie procedures taken from spiffy-utils and ported to WU
(define (wu:cookie-set! name value #!key (comment #f) (max-age #f) (domain #f) (path #f) (secure #f))
  (let ((conc-cond (lambda (s1 s2 s3) (if s2 (conc s1 s2 s3) ""))))
    (wu:set-header!
     "Set-Cookie"
     (string-append (->string name) "=\"" (->string value) "\""
		    (conc-cond "; Comment=\"" comment "\"")
		    (conc-cond "; Max-Age=\"" max-age "\"")
		    (conc-cond "; Domain=\"" domain "\"")
		    (conc-cond "; Path=\"" path "\"")
		    (if secure "; Secure" "")
		    "; Version=1"))))

;;; Delete a cookie by settings its maximum age to 0 seconds
(define (wu:cookie-delete! name #!key (domain #f) (path #f))
  (wu:cookie-set! (->string name) "" max-age: 0 domain: domain path: path))

; Get "cookie variable" value
; TODO: Implement cookie handling better we can get a cookie's Path/Domain
;	 and make the `$Version' cookie disappear to the user.
(define (get-request-cookies)
  (let ((cookie-hdr (wu:get-header "cookie")))
    (if cookie-hdr
	(remove null?
		(map (lambda (x)
		       ;; Each cookie is a 'name=value' pair.  Otherwise ignore it.
		       (match
			(string-match "[ \t]*([^ \t]+)[ \t]*=[ \t]*\"([^ \t]*)\"[ \t]*" x)
			((_ name value) (cons name value))
			(_ '())))
                                        ; RFC says we should accept `,' and `;' as cookie separators
		     (string-split cookie-hdr ",;")))
	(list))))

;; From HTTP egg. TODO: export it
(define (parse-encoded-arguments args)
  (let ([vals (string-split args "&")])
    (map (lambda (def)
	   (regex-case def
	     ["([^=]+)=(.*)" (_ name val)
	      (cons (http:canonicalize-string name) (http:canonicalize-string val))]
	     [else (cons (http:canonicalize-string def) "")] ) )
	 vals) ) )
